import os

from celery import Celery
from dagster_celery.config import CeleryConfig
from dagster_graphql.client.mutations import execute_execute_plan_mutation
from kombu import Queue

from dagster import ExecutionTargetHandle
from dagster.core.instance import InstanceRef
from dagster.core.serdes import serialize_dagster_namedtuple

DEFAULT_BROKER = 'pyamqp://guest@{hostname}:5672//'.format(
    hostname=os.getenv('DAGSTER_CELERY_BROKER_HOST', 'localhost')
)


def create_task(celery_app, **task_kwargs):
    @celery_app.task(bind=True, name='execute_query', **task_kwargs)
    def _execute_query(_self, handle_dict, variables, instance_ref_dict):
        instance_ref = InstanceRef.from_dict(instance_ref_dict)
        handle = ExecutionTargetHandle.from_dict(handle_dict)

        events = execute_execute_plan_mutation(
            handle=handle, variables=variables, instance_ref=instance_ref,
        )
        serialized_events = [serialize_dagster_namedtuple(event) for event in events]
        return serialized_events

    return _execute_query


def make_app(config=CeleryConfig()):
    app_ = Celery('dagster', **dict({'broker': DEFAULT_BROKER}, **(config._asdict())))
    app_.loader.import_module('celery.contrib.testing.tasks')
    app_.conf.task_queues = [
        Queue('dagster', routing_key='dagster.#', queue_arguments={'x-max-priority': 10})
    ]
    app_.conf.task_routes = {
        'execute_query': {'queue': 'dagster', 'routing_key': 'dagster.execute_query'}
    }
    app_.conf.task_queue_max_priority = 10
    app_.conf.task_default_priority = 5
    return app_


app = make_app()

execute_query = create_task(app)

if __name__ == '__main__':
    app.worker_main()



from dagster import Field, List, PermissiveDict, String
from dagster.core.definitions.executor import executor

from .config import CeleryConfig


@executor(
    name='celery',
    config={
        'broker': Field(
            String,
            is_optional=True,
            description=(
                'The URL of the Celery broker. Default: '
                '\'pyamqp://guest@{os.getenv(\'DAGSTER_CELERY_BROKER_HOST\','
                '\'localhost\')}//\'.'
            ),
        ),
        'backend': Field(
            String,
            is_optional=True,
            default_value='rpc://',
            description='The URL of the Celery results backend. Default: \'rpc://\'.',
        ),
        'include': Field(
            List[String], is_optional=True, description='List of modules every worker should import'
        ),
        'config_source': Field(
            PermissiveDict(), is_optional=True, description='Settings for the Celery app.'
        ),
    },
)
def celery_executor(init_context):
    '''Celery-based executor.

    The Celery executor exposes config settings for the underlying Celery app under
    the ``config_source`` key. This config corresponds to the "new lowercase settings" introduced
    in Celery version 4.0 and the object constructed from config will be passed to the
    :py:class:`celery.Celery` constructor as its ``config_source`` argument.
    (See https://docs.celeryproject.org/en/latest/userguide/configuration.html for details.)
    
    The executor also exposes the ``broker``, `backend`, and ``include`` arguments to the
    :py:class:`celery.Celery` constructor.

    In the most common case, you may want to modify the ``broker`` and ``backend`` (e.g., to use
    Redis  instead of RabbitMQ). We expect that ``celery_settings`` will be less frequently
    modified, but that when solid executions are especially fast or slow, or when there are
    different requirements around idempotence or retry, it will make sense to execute pipelines
    with variations on these settings.

    **Config**:
    .. code-block::

        {
            broker?: 'pyamqp://guest@localhost//',  # The URL of the Celery broker
            backend?: 'rpc://', # The URL of the Celery results backend
            include?: ['my_module'], # List of modules every worker should import
            celery_settings: {
                ... # Celery app config
            }
        }

    If you'd like to configure a celery executor in addition to the
    :py:class:`~dagster.default_executors`, you should add it to the ``executor_defs`` defined on a
    :py:class:`~dagster.ModeDefinition` as follows:

    .. code-block:: python

        from dagster import ModeDefinition, default_executors, pipeline
        from dagster_celery import celery_executor

        @pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [celery_executor])])
        def celery_enabled_pipeline():
            pass

    '''

    return CeleryConfig(**init_context.executor_config)



import time
from collections import defaultdict

from dagster import check
from dagster.core.engine.engine_base import Engine
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.core.serdes import deserialize_json_to_dagster_namedtuple

from .config import DEFAULT_PRIORITY, DEFAULT_QUEUE, CeleryConfig
from .tasks import create_task, make_app

TICK_SECONDS = 1


class CeleryEngine(Engine):
    @staticmethod
    def execute(pipeline_context, execution_plan):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)

        check.param_invariant(
            isinstance(pipeline_context.executor_config, CeleryConfig),
            'pipeline_context',
            'Expected executor_config to be CeleryConfig got {}'.format(
                pipeline_context.executor_config
            ),
        )

        celery_config = pipeline_context.executor_config

        pipeline_name = pipeline_context.pipeline_def.name

        handle_dict = pipeline_context.execution_target_handle.to_dict()

        instance_ref_dict = pipeline_context.instance.get_ref().to_dict()

        environment_dict = dict(pipeline_context.environment_dict, execution={'in_process': {}})

        mode = pipeline_context.mode_def.name

        run_id = pipeline_context.pipeline_run.run_id

        app = make_app(celery_config)

        pending_steps = execution_plan.execution_deps()

        task_signatures = {}  # Dict[step_key, celery.Signature]
        apply_kwargs = defaultdict(dict)  # Dict[step_key, Dict[str, Any]]

        sort_by_priority = lambda step_key: (-1 * apply_kwargs[step_key]['priority'])

        for step_key in execution_plan.step_keys_to_execute:
            step = execution_plan.get_step_by_key(step_key)
            priority = step.metadata.get('dagster-celery/priority', DEFAULT_PRIORITY)
            queue = step.metadata.get('dagster-celery/queue', DEFAULT_QUEUE)
            task = create_task(app)

            variables = {
                'executionParams': {
                    'selector': {'name': pipeline_name},
                    'environmentConfigData': environment_dict,
                    'mode': mode,
                    'executionMetadata': {'runId': run_id},
                    'stepKeys': [step_key],
                }
            }
            task_signatures[step_key] = task.si(handle_dict, variables, instance_ref_dict)
            apply_kwargs[step_key] = {
                'priority': priority,
                'queue': queue,
                'routing_key': '{queue}.execute_query'.format(queue=queue),
            }

        step_results = {}  # Dict[ExecutionStep, celery.AsyncResult]
        completed_steps = set({})  # Set[step_key]

        while pending_steps or step_results:

            results_to_pop = []
            for step_key, result in sorted(
                step_results.items(), key=lambda x: sort_by_priority(x[0])
            ):
                if result.ready():
                    try:
                        step_events = result.get()
                    except Exception:  # pylint: disable=broad-except
                        # We will want to do more to handle the exception here.. maybe subclass Task
                        # Certainly yield an engine or pipeline event
                        step_events = []
                    for step_event in step_events:
                        yield deserialize_json_to_dagster_namedtuple(step_event)
                    results_to_pop.append(step_key)
                    completed_steps.add(step_key)
            for step_key in results_to_pop:
                if step_key in step_results:
                    del step_results[step_key]

            pending_to_pop = []
            for step_key, requirements in pending_steps.items():
                if requirements.issubset(completed_steps):
                    pending_to_pop.append(step_key)

            # This is a slight refinement. If we have n workers idle and schedule m > n steps for
            # execution, the first n steps will be picked up by the idle workers in the order in
            # which they are scheduled (and the following m-n steps will be executed in priority
            # order, provided that it takes longer to execute a step than to schedule it). The test
            # case has m >> n to exhibit this behavior in the absence of this sort step.
            to_execute = sorted(pending_to_pop, key=sort_by_priority)
            for step_key in to_execute:
                step_results[step_key] = task_signatures[step_key].apply_async(
                    **apply_kwargs[step_key]
                )

            for step_key in pending_to_pop:
                if step_key in pending_steps:
                    del pending_steps[step_key]

            time.sleep(TICK_SECONDS)



import os
import subprocess
import uuid

import click
import yaml
from celery.utils.nodenames import default_nodename, host_format

from dagster import check, seven
from dagster.core.instance import DagsterInstance
from dagster.core.types.evaluator.evaluation import evaluate_config
from dagster.utils import load_yaml_from_path, mkdir_p
from dagster.utils.indenting_printer import IndentingPrinter

from .config import CeleryConfig
from .executor import celery_executor
from .tasks import make_app


def create_worker_cli_group():
    group = click.Group(name='worker')
    group.add_command(worker_start_command)
    group.add_command(worker_terminate_command)
    group.add_command(worker_list_command)
    return group


def get_config_value_from_yaml(yaml_path):
    parsed_yaml = load_yaml_from_path(yaml_path) or {}
    # # Would be better not to hardcode this path
    return parsed_yaml.get('execution', {}).get('celery', {}) or {}


def get_app(config_yaml, config_module, config_file):
    check.invariant(
        not (sum(int(bool(x)) for x in [config_module, config_file, config_yaml]) > 1),
        'Can only set one of --config-yaml/-y --config-module/-m or --config-file/-f',
    )

    if config_yaml:
        celery_config = CeleryConfig(**get_config_value_from_yaml(config_yaml))
    else:
        celery_config = CeleryConfig()

    app = make_app(celery_config)

    if config_module:
        app.config_from_object(config_module)

    if config_file:
        config_object = seven.import_module_from_path(
            os.path.basename(config_file)[:-3], config_file
        )
        app.config_from_object(config_object)

    return app


def get_worker_name(name=None):
    return (
        name + '@%h'
        if name is not None
        else 'dagster-{uniq}@%h'.format(uniq=str(uuid.uuid4())[-6:])
    )


def celery_worker_config_args(config_yaml, config_module, config_file):
    check.invariant(
        not (sum(int(bool(x)) for x in [config_module, config_file, config_yaml]) > 1),
        'Can only set one of --config-yaml/-y --config-module/-m or --config-file/-f',
    )
    if config_yaml:
        instance = DagsterInstance.get()
        config_type = celery_executor.config_field.config_type
        config_value = get_config_value_from_yaml(config_yaml)
        config_hash = hash(yaml.dump(config_value))

        config_dir = os.path.join(instance.root_directory, 'dagster_celery', 'config')
        mkdir_p(config_dir)
        config_path = os.path.join(config_dir, '{config_hash}.py'.format(config_hash=config_hash))
        if not os.path.exists(config_path):
            validated_config = evaluate_config(config_type, config_value).value
            with open(config_path, 'w') as fd:
                printer = IndentingPrinter(printer=fd.write)
                if 'broker' in validated_config:
                    printer.line(
                        'broker_url = {broker_url}'.format(
                            broker_url=str(validated_config['broker'])
                        )
                    )
                if 'backend' in validated_config:
                    printer.line(
                        'result_backend = {result_backend}'.format(
                            result_backend=str(validated_config['backend'])
                        )
                    )
                if 'config_source' in validated_config:
                    for key, value in validated_config['config_source'].items():
                        printer.line('{key} = {value}'.format(key=key, value=str(value)))
        # n.b. right now we don't attempt to clean up this cache, but it might make sense to delete
        # any files older than some time if there are more than some number of files present, etc.
        return ['--config', config_path]
    elif config_module or config_file:
        return ['--config', config_module or config_file]
    else:
        return []


def launch_background_worker(subprocess_args):
    return subprocess.Popen(subprocess_args + ['--detach', '--pidfile='], stdout=None, stderr=None)


@click.command(name='start')
@click.option('--name', '-n', type=click.STRING, default=None)
@click.option('--config-yaml', '-y', type=click.Path(exists=True), default=None)
@click.option('--config-module', '-m', type=click.STRING, default=None)
@click.option('--config-file', '-f', type=click.Path(exists=True), default=None)
@click.option('--queue', '-q', type=click.STRING, multiple=True)
@click.option('--background', '-d', is_flag=True)
@click.option('--includes', '-i', type=click.STRING, multiple=True)
@click.option('--loglevel', '-l', type=click.STRING, default='INFO')
def worker_start_command(
    name=None,
    config_yaml=None,
    config_module=None,
    config_file=None,
    background=None,
    queue=None,
    includes=None,
    loglevel=None,
):
    '''dagster-celery start'''

    loglevel_args = ['--loglevel', loglevel]

    if len(queue) > 4:
        check.failed(
            'Can\'t start a dagster_celery worker that listens on more than four queues, due to a '
            'bug in Celery 4.'
        )
    queue_args = []
    for q in queue:
        queue_args += ['-Q', q]

    includes_args = ['-I', ','.join(includes)] if includes else []

    config_args = celery_worker_config_args(config_yaml, config_module, config_file)
    subprocess_args = (
        ['celery', '-A', 'dagster_celery.tasks', 'worker', '--prefetch-multiplier=1']
        + config_args
        + loglevel_args
        + queue_args
        + includes_args
        + ['-n', get_worker_name(name)]
    )

    if background:
        launch_background_worker(subprocess_args)
    else:
        return subprocess.check_call(subprocess_args)


@click.command(name='list')
@click.option('--config-yaml', '-y', type=click.Path(exists=True), default=None)
@click.option('--config-module', '-m', type=click.STRING, default=None)
@click.option('--config-file', '-f', type=click.Path(exists=True), default=None)
def worker_list_command(config_yaml=None, config_module=None, config_file=None):
    app = get_app(config_yaml, config_module, config_file)

    print(app.control.inspect(timeout=1).active())


@click.command(name='terminate')
@click.argument('name', default='dagster')
@click.option('--all', '-a', 'all_', is_flag=True)
@click.option('--config-yaml', '-y', type=click.Path(exists=True), default=None)
@click.option('--config-module', '-m', type=click.STRING, default=None)
@click.option('--config-file', '-f', type=click.Path(exists=True), default=None)
def worker_terminate_command(
    name='dagster', config_yaml=None, config_module=None, config_file=None, all_=False
):
    app = get_app(config_yaml, config_module, config_file)

    if all_:
        app.control.broadcast('shutdown')
    else:
        app.control.broadcast(
            'shutdown', destination=[host_format(default_nodename(get_worker_name(name)))]
        )


worker_cli = create_worker_cli_group()


@click.group(commands={'worker': worker_cli})
def main():
    '''dagster-celery'''


if __name__ == '__main__':
    # pylint doesn't understand click
    main()  # pylint:disable=no-value-for-parameter
